/*------------- Telecommunications & Signal Processing Lab -------------
                           McGill University

Routine:
  void SPoptions (int argc, const char *argv[], struct SP_PlayOpt *PlayOpt,
                  struct SP_FIpar FI[MAXFILE],
		  struct AD_DevOpt DevOpt[MAXFILE], int *Nfiles)

Purpose:
  Decode options for SunPlay

Description:
  This routine decodes options for SunPlay.

Parameters:
   -> int argc
      Number of command line arguments
   -> const char *argv[]
      Array of pointers to argument strings
  <-  struct SP_PlayOpt *PlayOpt
      Playback options (device name, repeats, etc.)
  <-  struct SP_FIpar FI[MAXFILE]
      Input file parameters
  <-  struct AD_DevOpt DevOpt[MAXFILE]
      Audio device playback options
  <-  int *Nfiles
      Number of input file names (max MAXFILE)
 
Author / revision:
  P. Kabal  Copyright (C) 1999
  $Revision: 1.19 $  $Date: 1999/06/14 16:28:52 $

----------------------------------------------------------------------*/

static char rcsid[] = "$Id: SPoptions.c 1.19 1999/06/14 AFsp-v5r0 $";

#include <stdio.h>		/* FILENAME_MAX */
#include <assert.h>

#include <libtsp.h>
#include <AD.h>
#include <AO.h>
#include "SunPlay.h"

#define ERRSTOP(text,par)	UThalt ("%s: %s: \"%s\"", PROGRAM, text, par)

/* Option table */
static const char *OptTable[] = {
  "-r#", "--r*epeat=",
  "-B#", "--pau*se=",
  "-d#", "--d*evice=",
  "-w",  "--w*ait",
  "-nw", "--now*ait",
  NULL
};


void
SPoptions (int argc, const char *argv[], struct SP_PlayOpt *PlayOpt,
	   struct SP_FIpar FI[], struct AD_DevOpt DevOpt[], int *Nfiles)

{
  struct SP_FIpar FIx;
  struct AD_DevOpt DevOptx;
  const char *OptArg;

  int nF, n, N;
  int FIParSet, DevOptSet;

/* Input file defaults */
  FIParSet = 0;
  FIpar_INIT (&FIx);

/* Output port defaults */
  DevOptSet = 0;
  DevOpt_INIT (&DevOptx);

/* Program option defaults */
  PlayOpt_INIT (PlayOpt);

/* Initialization */
  UTsetProg (PROGRAM);
  nF = 0;

/* Decode options */
  AOinitOpt (argc, argv);
  while (1) {

    /* Decode input file options */
    n = AOdecFI (&FIx);
    if (n >= 1) {
      FIParSet = nF;
      continue;
    }

    /* Decode audio port options */
    n = ADdecDevOpt (&DevOptx);
    if (n >= 1) {
      DevOptSet = nF;
      continue;
    }

    /* Decode help options */
    n = AOdecHelp (VERSION, SPMF_Usage);
    if (n >= 1)
      continue;

    /* Decode program options */
    n = AOdecOpt (OptTable, &OptArg);
    if (n == -1)
      break;

    switch (n) {
    case -2:
      UThalt (SPMF_Usage, PROGRAM);
      break;

    case 0:
      /* Filename argument */
      ++nF;
      if (nF > MAXFILE)
	UThalt ("%s: %s", PROGRAM, SPM_XFName);
      STcopyMax (OptArg, FIx.Fname, FILENAME_MAX);
      FI[nF-1] = FIx;
      DevOpt[nF-1] = DevOptx;
      break;
    case 1:
    case 2:
      /* Number of repetitions */
      if (STdec1int (OptArg, &PlayOpt->Nrepeat) || PlayOpt->Nrepeat <= 0)
	ERRSTOP (SPM_BadRep, OptArg);
      break;
    case 3:
    case 4:
      /* Pauses */
      if (STdecNfloat (OptArg, 1, 2, PlayOpt->Pause, &N) ||
	  PlayOpt->Pause[0] < 0.0 || (N == 2 && PlayOpt->Pause[1] < 0.0))
	ERRSTOP (SPM_BadPause, OptArg);
      if (N == 1)
	PlayOpt->Pause[1] = PlayOpt->Pause[0];
      break;
    case 5:
    case 6:
      /* Audio device */
      PlayOpt->AudioDev = OptArg;
      break;
    case 7:
    case 8:
      /* Wait flag */
      PlayOpt->Wait_flag = AD_WAIT_FLAG;
      break;
    case 9:
    case 10:
      /* Nowait */
      PlayOpt->Wait_flag = 0;
      break;
    default:
      assert (0);
      break;
    }
  }

/* Checks */
  if (nF <= 0)
    UThalt ("%s: %s", PROGRAM, SPM_NoFName);
  if (FIParSet >= nF)
    UThalt ("%s: %s", PROGRAM, SPM_LateFPar);
  if (DevOptSet >= nF)
    UThalt ("%s: %s", PROGRAM, SPM_LateDevPar);

  /* Check for too many stdin specs */
  AOstdin (FI, nF);

/* Return values */
  *Nfiles = nF;

  return;
}
